import { valid, validRange, satisfies } from 'semver';

var ResolutionType;
(function (ResolutionType) {
    ResolutionType["Available"] = "available";
    ResolutionType["LoadFrom"] = "load-from";
})(ResolutionType || (ResolutionType = {}));

var NPMURLPattern = /^(?:npm:\/\/)?((?:@[\w\.\-]+\/)?\w[\w\.\-]+)(@[^\/]+)?(\/.*)?$/;
function splitNPMURL(url) {
    var loaders = extractLoaders(url);
    var requestWithoutLoaders = removeLoaders(url);
    var match = requestWithoutLoaders.match(NPMURLPattern);
    if (match) {
        var _ = match[0], name_1 = match[1], version = match[2], pathname = match[3];
        return {
            loaders: loaders,
            name: name_1,
            version: version,
            pathname: pathname,
        };
    }
}
var DefaultResolver = /** @class */ (function () {
    function DefaultResolver() {
        // Maps known URLs to their module id.
        this.knownURLs = {};
        // Records redirects due to extensions, checking index.js, the package.json's
        // `main` field, etc.
        this.npmPathRedirects = {};
        // For each package, store a list of versions that we have any ids for.
        this.npmKnownPackageVersions = {};
        // Keep track of version aliases of each package, so that if an alias is
        // referenced in multiple places, it can be immediately resolved to the
        // appropriate actual version.
        this.npmVersionAliases = {};
        // Keep track of the first used version for each package, and use that for all
        // subsequent bare references to the same package.
        this.npmDefaultVersions = {};
    }
    DefaultResolver.prototype.registerId = function (id) {
        this.knownURLs[id] = id;
    };
    // Register an id along with the URL that requested it, to speed up future
    // resolutions.
    DefaultResolver.prototype.registerResolvedURL = function (url, id) {
        this.knownURLs[url] = id;
        var idNPMMatch = removeLoaders(id).match(NPMURLPattern);
        if (idNPMMatch) {
            var _ = idNPMMatch[0], name_2 = idNPMMatch[1], version = idNPMMatch[2], _a = idNPMMatch[3], path = _a === void 0 ? '' : _a;
            var urlNPMMatch = removeLoaders(url).match(NPMURLPattern);
            if (urlNPMMatch) {
                var urlNPMVersion = urlNPMMatch[2] || '';
                if (urlNPMVersion !== version) {
                    // We've received an actual version in response to an alias or range.
                    this.npmVersionAliases[urlNPMVersion] = version;
                }
                var urlNPMPath = urlNPMMatch[3] || '';
                if (urlNPMPath !== path) {
                    this.npmPathRedirects[name_2 + version + urlNPMPath] = path;
                }
            }
            // For each package, store a list of versions that we have any ids for,
            // so that we can match ranges against known versions.
            if (!this.npmKnownPackageVersions[name_2]) {
                this.npmKnownPackageVersions[name_2] = [version];
            }
            else {
                this.npmKnownPackageVersions[name_2].push(version);
            }
        }
    };
    // Either returns an { id } object containing a registered module id, or a
    // { url } object containing a URL that can be resolved to a module id/source
    // by the breadboard host.
    //
    // ids are all urls, but urls aren't necessarily ids.
    DefaultResolver.prototype.resolve = function (request, parentId, packageDependencies) {
        var url;
        var loaders = extractLoaders(request);
        var requestWithoutLoaders = removeLoaders(request);
        var npmMatch = requestWithoutLoaders.match(NPMURLPattern);
        if (!npmMatch) {
            // It's not bare, so resolve the URL relative to the parent
            // and then try to match it against our npm pattern again.
            url = resolveIfNotPlainOrUrl(requestWithoutLoaders, parentId);
            npmMatch = url.match(NPMURLPattern);
        }
        var name, version, path;
        if (npmMatch) {
            name = npmMatch[1];
            version = npmMatch[2];
            path = npmMatch[3] || '';
            // If we got no version or a version alias, try and match it to a
            // real version.
            if (!version) {
                var packageVersion = packageDependencies[name];
                if (packageVersion) {
                    packageVersion = '@' + packageDependencies[name];
                }
                version = packageVersion || this.npmDefaultVersions[name] || '@latest';
            }
            if (this.npmVersionAliases[name + version]) {
                version = this.npmVersionAliases[name + version];
            }
            else if (!valid(version.slice(1))) {
                version = this.getKnownMatchingVersion(name, version) || version;
            }
            // If this is the first time we've seen this package, store the version
            // we're using for future versionless references.
            if (!this.npmDefaultVersions[name]) {
                this.npmDefaultVersions[name] = version;
            }
            path = this.npmPathRedirects[name + version + path] || path;
            url = 'npm://' + name + version + path;
        }
        // If we've seen this URL before, immediately return the corresponding id.
        var id = this.knownURLs[url];
        if (id) {
            return { type: ResolutionType.Available, id: loaders + id };
        }
        return { type: ResolutionType.LoadFrom, url: loaders + url };
    };
    DefaultResolver.prototype.getKnownMatchingVersion = function (name, range) {
        var rangeWithoutSymbol = range.slice(1);
        if (validRange(rangeWithoutSymbol)) {
            var versions = this.npmKnownPackageVersions[name] || [];
            for (var i = 0; i < versions.length; i++) {
                var version = versions[i];
                if (satisfies(version.slice(1), rangeWithoutSymbol)) {
                    return version;
                }
            }
        }
    };
    return DefaultResolver;
}());
/**
 * From system.js
 * Copyright (C) 2013-2018 Guy Bedford
 * https://github.com/systemjs/systemjs
 */
var backslashRegEx = /\\/g;
function resolveIfNotPlainOrUrl(relUrl, parentUrl) {
    if (relUrl.indexOf('://') >= 0) {
        return relUrl;
    }
    if (relUrl.indexOf('\\') !== -1)
        relUrl = relUrl.replace(backslashRegEx, '/');
    // protocol-relative
    if (relUrl[0] === '/' && relUrl[1] === '/') {
        return parentUrl.slice(0, parentUrl.indexOf(':') + 1) + relUrl;
    }
    // relative-url
    else if (relUrl[0] === '.' && (relUrl[1] === '/' || relUrl[1] === '.' && (relUrl[2] === '/' || relUrl.length === 2 && (relUrl += '/')) ||
        relUrl.length === 1 && (relUrl += '/')) ||
        relUrl[0] === '/') {
        var parentProtocol = parentUrl.slice(0, parentUrl.indexOf(':') + 1);
        // Disabled, but these cases will give inconsistent results for deep backtracking
        //if (parentUrl[parentProtocol.length] !== '/')
        //  throw new Error('Cannot resolve');
        // read pathname from parent URL
        // pathname taken to be part after leading "/"
        var pathname = void 0;
        if (parentUrl[parentProtocol.length + 1] === '/') {
            // resolving to a :// so we need to read out the auth and host
            if (parentProtocol !== 'file:') {
                pathname = parentUrl.slice(parentProtocol.length + 2);
                pathname = pathname.slice(pathname.indexOf('/') + 1);
            }
            else {
                pathname = parentUrl.slice(8);
            }
        }
        else {
            // resolving to :/ so pathname is the /... part
            pathname = parentUrl.slice(parentProtocol.length + (parentUrl[parentProtocol.length] === '/'));
        }
        if (relUrl[0] === '/')
            return parentUrl.slice(0, parentUrl.length - pathname.length - 1) + relUrl;
        // join together and split for removal of .. and . segments
        // looping the string instead of anything fancy for perf reasons
        // '../../../../../z' resolved to 'x/y' is just 'z'
        var segmented = pathname.slice(0, pathname.lastIndexOf('/') + 1) + relUrl;
        var output = [];
        var segmentIndex = -1;
        for (var i = 0; i < segmented.length; i++) {
            // busy reading a segment - only terminate on '/'
            if (segmentIndex !== -1) {
                if (segmented[i] === '/') {
                    output.push(segmented.slice(segmentIndex, i + 1));
                    segmentIndex = -1;
                }
            }
            // new segment - check if it is relative
            else if (segmented[i] === '.') {
                // ../ segment
                if (segmented[i + 1] === '.' && (segmented[i + 2] === '/' || i + 2 === segmented.length)) {
                    output.pop();
                    i += 2;
                }
                // ./ segment
                else if (segmented[i + 1] === '/' || i + 1 === segmented.length) {
                    i += 1;
                }
                else {
                    // the start of a new segment as below
                    segmentIndex = i;
                }
            }
            // it is the start of a new segment
            else {
                segmentIndex = i;
            }
        }
        // finish reading out the last segment
        if (segmentIndex !== -1)
            output.push(segmented.slice(segmentIndex));
        return parentUrl.slice(0, parentUrl.length - pathname.length) + output.join('');
    }
}
function extractLoaders(url) {
    var match = url.match(/^(.*!)*/);
    return match ? match[0] : '';
}
function removeLoaders(url) {
    return url.replace(/^(.*!)*/, '');
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var PolestarError = /** @class */ (function (_super) {
    __extends(PolestarError, _super);
    function PolestarError() {
        var _newTarget = this.constructor;
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var _this = _super.apply(this, args) || this;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.name = _newTarget.name;
        // Maintains proper stack trace for where our error was thrown (only available on V8)
        if (Error.captureStackTrace) {
            Error.captureStackTrace(_this, _newTarget);
        }
        return _this;
    }
    return PolestarError;
}(Error));
var UnresolvableError = /** @class */ (function (_super) {
    __extends(UnresolvableError, _super);
    function UnresolvableError(request, fromModuleId) {
        var _newTarget = this.constructor;
        var _this = _super.call(this, "Module \"" + request + "\" could not be resolved by require.resolve from module \"" + fromModuleId + "\".") || this;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.request = request;
        _this.fromModuleId = fromModuleId;
        return _this;
    }
    return UnresolvableError;
}(PolestarError));
var CyclicDependencyError = /** @class */ (function (_super) {
    __extends(CyclicDependencyError, _super);
    function CyclicDependencyError(request) {
        var _this = _super.call(this, "Module \"" + request + "\" can not be required from itself.") || this;
        _this.request = request;
        return _this;
    }
    return CyclicDependencyError;
}(PolestarError));

var nextScriptId = 1;
var scriptIdPrefix = Math.random().toString(36).slice(2);
var __polestar_script__ = function (id, fn) {
    __polestar_script__[id] = fn;
};
if (!window['__polestar_script__']) {
    window['__polestar_script__'] = __polestar_script__;
}
else {
    throw new Error("Cannot load multiple copies of Polestar.");
}
// Running code with `eval` will cause some errors messages to be hidden from
// the script due to Chrome treating them as originating from another origin.
// Evaluating code by running it in script tags is supposedly slower, but
// allows errors to be properly forwarded to non-native consoles.
function createFunction(argNames, source) {
    var scriptId = scriptIdPrefix + '/' + nextScriptId++;
    var script = document.createElement('script');
    script.innerHTML =
        "(function(){__polestar_script__(" + JSON.stringify(scriptId) + ", function module(" + argNames.join(',') + ") {" +
            source +
            "\n})})()";
    // Script tags with inline source are synchronously executed when appended
    // to the document body.
    document.body.appendChild(script);
    var result = __polestar_script__[scriptId];
    delete __polestar_script__[scriptId];
    document.body.removeChild(script);
    return result;
}

function createPolestar(options) {
    if (options === void 0) { options = {}; }
    return new Polestar(options);
}
var Polestar = /** @class */ (function () {
    function Polestar(options) {
        var _this = this;
        this.hasCalledOnEntry = false;
        this.nextEntryId = 1;
        this.handleFetchResult = function (result) {
            if (_this.error) {
                return Promise.reject(_this.error);
            }
            var url = result.url;
            var id = result.id;
            var load = _this.loads[url];
            _this.resolver.registerResolvedURL(url, id);
            // It's possible that this id has already been loaded using a different url.
            // If so, we need to let it know of any more requesting modules, in case
            // they're needed to avoid deadlock.
            var moduleWrapper = _this.moduleWrappers[id];
            if (moduleWrapper) {
                moduleWrapper.addToRequiredBy(load.requiredBy);
                return Promise.resolve(moduleWrapper);
            }
            return _this.prepareModuleWrapper(id, result.code, result.dependencies || [], result.dependencyVersionRanges || {}, load.requiredBy, result.css);
        };
        this.handleUnrequiredPrepared = function (moduleWrapper) {
            if (!_this.hasCalledOnEntry) {
                _this.options.onEntry();
            }
            moduleWrapper.execute();
            return moduleWrapper;
        };
        // If we encounter any error while loading/executing modules, stop loading
        // subsequent modules and notify our owner.
        this.setError = function (error) {
            var lastError = _this.error;
            _this.error = error;
            if (lastError !== error) {
                if (_this.options.onError) {
                    _this.options.onError(error);
                }
                else {
                    console.error("Error encountered while loading module:", error);
                }
            }
        };
        this.options = Object.assign({ globals: {}, onEntry: function () { } }, options);
        this.loads = {};
        this.moduleWrappers = {};
        this.resolver = options.resolver || new DefaultResolver();
    }
    Polestar.prototype.evaluate = function (dependencies, code, dependencyVersionRanges, id) {
        if (id === undefined) {
            id = 'polestar-anonymous://' + String(this.nextEntryId++);
        }
        return this.prepareModuleWrapper(id, code, dependencies || [], dependencyVersionRanges || {}).then(getModule);
    };
    Polestar.prototype.require = function (request) {
        var resolution = this.resolve(request);
        var url = resolution.type === ResolutionType.Available ? resolution.id : resolution.url;
        var modulePromise = this.loadWrapper(url).then(getModule);
        modulePromise.catch(this.setError);
        return modulePromise;
    };
    Polestar.prototype.resolve = function (request, parentId, defaultPackageVersionRanges) {
        if (defaultPackageVersionRanges === void 0) { defaultPackageVersionRanges = {}; }
        return this.resolver.resolve(request, parentId, defaultPackageVersionRanges);
    };
    // Initializes a module if it doesn't already exist
    Polestar.prototype.loadWrapper = function (url, requiredBy, originalRequest) {
        var moduleWrapper = this.moduleWrappers[url];
        if (moduleWrapper) {
            return Promise.resolve(moduleWrapper);
        }
        var existingLoad = this.loads[url];
        if (existingLoad) {
            if (requiredBy) {
                existingLoad.requiredBy.push(requiredBy);
            }
            return existingLoad.preparedModuleWrapperPromise;
        }
        var preparedModuleWrapperPromise = this.options.fetcher(url, {
            requiredById: requiredBy && requiredBy.module.id,
            originalRequest: originalRequest,
        })
            .then(this.handleFetchResult);
        this.loads[url] = {
            requiredBy: requiredBy ? [requiredBy] : [],
            preparedModuleWrapperPromise: preparedModuleWrapperPromise,
        };
        return preparedModuleWrapperPromise;
    };
    Polestar.prototype.prepareModuleWrapper = function (id, code, dependencies, dependencyVersionRanges, requiredBy, css) {
        var _this = this;
        if (requiredBy === void 0) { requiredBy = []; }
        if (this.error) {
            return Promise.reject(this.error);
        }
        if (css) {
            var head = document.head || document.getElementsByTagName('head')[0];
            var style = document.createElement('style');
            style.type = 'text/css';
            style.appendChild(document.createTextNode(css));
            head.appendChild(style);
        }
        try {
            var moduleWrapper_1;
            var globals = this.options.globals || {};
            var globalNames = Object.keys(globals);
            var globalValues = [];
            for (var i = 0; i < globalNames.length; i++) {
                globalValues.push(globals[globalNames[i]]);
            }
            var prepareDependencies_1;
            // For UMD modules, we'll use their dependencies instead of data.prepareRequires.
            if (dependencies === 'umd') {
                var define = function (name, dependencies, factory) {
                    // name and dependencies are optional
                    if (!factory) {
                        factory = dependencies;
                        dependencies = Array.isArray(name) ? name : [];
                        if (!factory) {
                            factory = name;
                            dependencies = [];
                        }
                    }
                    // Some UMD packages depend on an "exports" object, which shouldn't
                    // be required like normal objects.
                    prepareDependencies_1 = dependencies.slice(0);
                    var index = prepareDependencies_1.indexOf('exports');
                    if (index !== -1) {
                        prepareDependencies_1.splice(index, 1);
                    }
                    moduleWrapper_1 = new ModuleWrapper(_this, id, dependencyVersionRanges, function (require, module, exports) {
                        var factoryResult = factory.apply(null, dependencies.map(function (dep) { return dep === 'exports' ? exports : require(dep); }));
                        if (factoryResult !== undefined) {
                            module.exports = factoryResult;
                        }
                    });
                };
                define.amd = true;
                var fn = createFunction(['define'].concat(globalNames), code);
                var args = [define].concat(globalValues);
                fn.apply(this.options.moduleThis, args);
            }
            else {
                globalNames.push('require', 'module', 'exports');
                prepareDependencies_1 = dependencies;
                var moduleFunction = createFunction(globalNames, code);
                var boundModuleFunction = Function.prototype.bind.apply(moduleFunction, [this.options.moduleThis].concat(globalValues));
                moduleWrapper_1 = new ModuleWrapper(this, id, dependencyVersionRanges, boundModuleFunction);
            }
            this.resolver.registerId(id);
            this.moduleWrappers[id] = moduleWrapper_1;
            var preparedPromise = moduleWrapper_1.prepare(prepareDependencies_1, requiredBy);
            if (requiredBy.length === 0) {
                preparedPromise.catch(this.setError);
                preparedPromise = preparedPromise.then(this.handleUnrequiredPrepared);
                preparedPromise.catch(this.setError);
            }
            return preparedPromise;
        }
        catch (error) {
            this.setError(error);
            return Promise.reject(error);
        }
    };
    return Polestar;
}());
var ModuleWrapper = /** @class */ (function () {
    function ModuleWrapper(loader, id, dependencyVersionRanges, fn) {
        var _this = this;
        this.stopWaitingFor = function (url) {
            var index = _this.waitingFor.indexOf(url);
            if (index !== -1) {
                _this.waitingFor.splice(index, 1);
                if (_this.waitingFor.length === 0) {
                    _this.isPrepared = true;
                    _this.resolvePrepared(_this);
                }
            }
        };
        this.addRequiredBy = function (moduleWrapper) {
            _this.requiredBy.add(moduleWrapper);
        };
        this.requiredBy = new Set();
        this.waitingFor = [];
        this.fn = fn;
        this.polestar = loader;
        this.dependencyVersionRanges = dependencyVersionRanges;
        this.isPrepared;
        this.preparedPromise = new Promise(function (resolve, reject) {
            _this.resolvePrepared = resolve;
            _this.rejectPrepared = reject;
        });
        var require = (function (request) {
            var requestedId = require.resolve(request);
            if (id === requestedId) {
                throw new CyclicDependencyError(id);
            }
            var requestedWrapper = loader.moduleWrappers[requestedId];
            if (!requestedWrapper.module.loaded) {
                requestedWrapper.execute();
            }
            return requestedWrapper.module.exports;
        });
        require.resolve = function (request) {
            var result = _this.polestar.resolve(request, id, dependencyVersionRanges);
            if (result.type !== ResolutionType.Available) {
                throw new UnresolvableError(request, id);
            }
            return result.id;
        };
        this.module = {
            exports: {},
            id: id,
            loaded: false,
            require: require,
        };
    }
    // Actually execute the module code
    ModuleWrapper.prototype.execute = function () {
        var module = this.module;
        // This should never happen, but just in case...
        if (module.loaded) {
            throw new Error("Can't execute module " + this.module.id + " twice.");
        }
        module.loaded = true;
        this.fn(module.require, module, module.exports);
    };
    // This is separate from the constructor, as it allows the loader to
    // register this module under multiple URLs before loading/preparing any
    // of the module's dependencies.
    ModuleWrapper.prototype.prepare = function (dependencyRequests, requiredByWrappers) {
        var _this = this;
        if (dependencyRequests === void 0) { dependencyRequests = []; }
        if (requiredByWrappers === void 0) { requiredByWrappers = []; }
        // Build a set of all all modules depending on us, modules depending on
        // modules depending on us, etc.
        for (var i = 0; i < requiredByWrappers.length; i++) {
            var requiredBy = requiredByWrappers[i];
            this.addRequiredBy(requiredBy);
            requiredBy.requiredBy.forEach(this.addRequiredBy);
        }
        var _loop_1 = function (i) {
            var request = dependencyRequests[i];
            if (request) {
                var result = this_1.polestar.resolve(request, this_1.module.id, this_1.dependencyVersionRanges);
                if (result.type === ResolutionType.Available) {
                    var dependencyWrapper = this_1.polestar.moduleWrappers[result.id];
                    // If the requested module depends on us, then waiting for the
                    // requested module to be ready will result in deadlock.
                    if (!this_1.requiredBy.has(dependencyWrapper) && !dependencyWrapper.isPrepared) {
                        this_1.waitingFor.push(result.id);
                        dependencyWrapper.preparedPromise
                            .then(getWrapperId)
                            .then(this_1.stopWaitingFor)
                            .catch(this_1.rejectPrepared);
                    }
                }
                else {
                    var url_1 = result.url;
                    this_1.waitingFor.push(result.url);
                    this_1.polestar.loadWrapper(url_1, this_1, request)
                        .then(function () { return _this.stopWaitingFor(url_1); })
                        .catch(this_1.rejectPrepared);
                }
            }
        };
        var this_1 = this;
        for (var i = 0; i < dependencyRequests.length; i++) {
            _loop_1(i);
        }
        if (!this.waitingFor.length) {
            this.isPrepared = true;
            this.resolvePrepared(this);
        }
        return this.preparedPromise;
    };
    ModuleWrapper.prototype.addToRequiredBy = function (requiredByWrappers) {
        // Update our list of modules that are waiting for us, any modules
        // that are waiting for modules that are waiting for us, etc.
        for (var i = 0; i < requiredByWrappers.length; i++) {
            var requiredBy = requiredByWrappers[i];
            this.addRequiredBy(requiredBy);
            requiredBy.requiredBy.forEach(this.addRequiredBy);
        }
        // Stop waiting for any modules that are waiting for us, any modules
        // that are waiting for modules that are waiting for us, etc.
        for (var i = 0; i < this.waitingFor.length; i++) {
            var waitingForURL = this.waitingFor[i];
            var result = this.polestar.resolve(waitingForURL, this.module.id, this.dependencyVersionRanges);
            // It can't be waiting for us if it hasn't loaded yet
            if (result.type === ResolutionType.Available && this.requiredBy.has(this.polestar.moduleWrappers[result.id])) {
                this.stopWaitingFor(waitingForURL);
            }
        }
    };
    return ModuleWrapper;
}());
function getWrapperId(moduleWrapper) {
    return moduleWrapper.module.id;
}
function getModule(moduleWrapper) {
    return moduleWrapper.module;
}

export { splitNPMURL, DefaultResolver, PolestarError, UnresolvableError, CyclicDependencyError, createPolestar, Polestar, ModuleWrapper, ResolutionType };
//# sourceMappingURL=polestar.js.map
